/*
 * Copyright (C) 1995, jack
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

static char rcsid[]= "$Id: makeargv.c,v 1.2 1995/11/09 08:40:50 jack Exp $";

#include <ctype.h>
#include <stdlib.h>

#include "config.h"
#include "makeargv.h"
#include "utils.h"

#define USE_MULTIBYTE 1

#define ADD_UNITS 8
#define NO_QUOTE 0
#define SINGLE_QUOTE 001
#define DOUBLE_QUOTE 002
#define BACKSLASH_QUOTE 004
#ifdef USE_MULTIBYTE
#  define MULTIBYTE_QUOTE 010
#  define ISMBFIRST(c) ((((c) >= 0x80) && ((c) <= 0x9f)) \
			|| (((c) >= 0xe0) && ((c) <= 0xff)))
#endif

char **
makeargv (char *string)
{
  char **argv = NULL;
  char *argbuf;
  char *arg;
  int argc = 0;
  int maxargc = 0;
  int quote = NO_QUOTE, bsquote = NO_QUOTE;

  if (string != NULL)
    {
      while (isspace (*string))
	string++;
      if (*string == '\0')
	return NULL;

      argbuf = xstrdup (string);
      do
	{
	  arg = argbuf;
	  while (*string)
	    {
	      if (bsquote != NO_QUOTE)
		{
		  bsquote = 0;
		  *arg++ = *string++;
		  continue;
		}
#ifdef USE_MULTIBYTE
	      else if (ISMBFIRST ((unsigned char) *string))
		{
		  bsquote = MULTIBYTE_QUOTE;
		  *arg++ = *string++;
		  continue;
		}
#endif

	      if (quote == SINGLE_QUOTE)
		{
		  if (*string == '\'')
		    {
		      quote = NO_QUOTE;
		      string++;
		      continue;
		    }
		}
	      else if (*string == '\\')
		{
		  bsquote = BACKSLASH_QUOTE;
		  string++;
		  continue;
		}
	      else if (quote == DOUBLE_QUOTE)
		{
		  if (*string == '"')
		    {
		      quote = NO_QUOTE;
		      string++;
		      continue;
		    }
		}
	      else
		{
		  if (isspace (*string))
		    break;
		  else if (*string == '\'')
		    {
		      quote = SINGLE_QUOTE;
		      string++;
		      continue;
		    }
		  else if (*string == '"')
		    {
		      quote = DOUBLE_QUOTE;
		      string++;
		      continue;
		    }
		}
	      *arg++ = *string++;
	    }
	  *arg++ = '\0';

	  if (bsquote != NO_QUOTE || quote != NO_QUOTE)
	    fatal ("Unterminated string");

	  if ((argc + 1) >= maxargc)
	    {
	      maxargc += ADD_UNITS;
	      argv = (char **) xrealloc (argv, sizeof (*argv) * maxargc);
	    }
	  argv[argc++] = xstrdup (argbuf);

	  while (isspace (*string))
	    string++;
	}
      while (*string != '\0');
      free (argbuf);
      argv[argc++] = NULL;
    }

  return argv;
}

char **
copyargv (char **argv)
{
  char **newargv = NULL;
  int newargc;
  int argc;

  if (argv != NULL)
    {
      argc = countargv (argv);
      newargv = (char **) xmalloc (sizeof (*argv) * (argc + 1));
      for (newargc = 0; newargc < argc; newargc++)
	newargv[newargc] = xstrdup (argv[newargc]);
      newargv[newargc] = NULL;
    }

  return newargv;
}

void
freeargv (char **argv)
{
  int argc = 0;

  if (argv != NULL)
    {
      while (argv[argc] != NULL)
	{
	  free (argv[argc]);
	  argc++;
	}
      free (argv);
    }
}

char **
addargv (char **baseargv, char **argv)
{
  char **newargv = NULL;
  int newargc;
  int baseargc;
  int argc;

  if (baseargv == NULL)
    newargv = copyargv (argv);
  else if (argv == NULL)
    newargv = copyargv (baseargv);
  else
    {
      baseargc = countargv (baseargv);
      argc = countargv (argv);
      newargv = (char **) xmalloc (sizeof (*argv) * (baseargc + argc + 1));
      for (newargc = 0; newargc < baseargc; newargc++)
	newargv[newargc] = xstrdup (baseargv[newargc]);
      for (newargc = baseargc; newargc < (baseargc + argc); newargc++)
	newargv[newargc] = xstrdup (argv[newargc - baseargc]);
      newargv[newargc] = NULL;
    }

  return newargv;
}

char **
insertargv (char **baseargv, char **argv)
{
  char **newargv = NULL;
  int newargc;
  int baseargc;
  int argc;
  int i;

  if (baseargv == NULL)
    newargv = NULL;
  else if (argv == NULL)
    newargv = copyargv (baseargv);
  else
    {
      baseargc = countargv (baseargv);
      argc = countargv (argv);
      newargv = (char **) xmalloc (sizeof (*argv) * (baseargc + argc + 1));

      /* fixed argv[0] */
      newargc = 0;
      newargv[newargc++] = xstrdup (baseargv[0]);

      for (i = 0; i < argc; i++)
	newargv[newargc++] = xstrdup (argv[i]);
      for (i = 1; i < baseargc; i++)
	newargv[newargc++] = xstrdup (baseargv[i]);
      newargv[newargc] = NULL;
    }

  return newargv;
}

int
countargv (char **argv)
{
  int argc = 0;

  if (argv == NULL)
    argc = -1;
  else
    {
      while (argv[argc] != NULL)
	argc++;
    }

  return argc;
}
