/*
 * Copyright (C) 1995, jack
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

static char rcsid[]= "$Id: utils.c,v 1.2 1995/10/21 12:19:54 jack Exp $";

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "config.h"
#include "utils.h"

extern char *progname;

char *
toslash (char *fname)
{
  char *p = fname;
  char c;

  for (c = *p; c != '\0'; c = *++p)
    if (c == '\\')
      *p = '/';

  return fname;
}

char *
get_basename (char *fname)
{
  char *p;

#if defined (__HUMAN68K__) || defined (__MSDOS__)
  if ((p = strrchr (fname, ':')) != NULL)
    fname = p + 1;
  if ((p = strrchr (fname, '\\')) != NULL)
    fname = p + 1;
#endif
  if ((p = strrchr (fname, '/')) != NULL)
    fname = p + 1;

  return fname;
}

char *
suppress_extension (char *fname)
{
  int len;

#if defined (__HUMAN68K__)
  if ((len = strlen (fname)) > 2)
    if (fname[len - 2] == '.')
      switch (fname[len - 1])
	{
	case 'x': case 'X':
	case 'r': case 'R':
	case 'z': case 'Z':
	  fname[len - 2] = '\0';
	  break;
	}
#elif defined (__MSDOS__)
  if ((len = strlen (fname)) > 4)
    if (fname[len - 4] == '.')
      if ((! stricmp (fname + len - 3, "EXE"))
	  || (! stricmp (fname + len - 3, "COM")))
	fname[len - 4] = '\0';
#endif

  return fname;
}

#if !defined (HAVE_STRICMP) && !defined (HAVE_STRCASECMP)
int
stricmp (char *s1, char *s2)
{
  unsigned char *p1 = (unsigned char *)s1;
  unsigned char *p2 = (unsigned char *)s2;
  unsigned char c1, c2;

  if (p1 == p2)
    return 0;

  do
    {
      c1 = tolwer (*s1);
      c2 = tolwer (*s2);
      if (c1 == '\0')
	break;
      s1++;
      s2++;
    }
  while (c1 == c2);

  return (int)(c1 - c2);
}
#endif

#if !defined (HAVE_STRNICMP) && !defined (HAVE_STRNCASECMP)
int
strnicmp (char *s1, char *s2, int n)
{
  unsigned char *p1 = (unsigned char *)s1;
  unsigned char *p2 = (unsigned char *)s2;
  unsigned char c1, c2;

  if (p1 == p2 || n == 0)
    return 0;

  do
    {
      c1 = tolwer (*s1);
      c2 = tolwer (*s2);
      if (c1 == '\0' || c1 != c2)
	break;
      s1++;
      s2++;
    }
  while (--n);

  return (int)(c1 - c2);
}
#endif

#if !defined (HAVE_STRLWR)
char *
strlwr (char *s)
{
  unsigned char *p =  (unsigned char *)s;
  unsigned char uc;

  while ((uc = *p))
    *p++ = tolower (uc);

  return s;
}
#endif

#if !defined (HAVE_STRDUP)
char *
strdup (char *s)
{
  char *p;

  p = malloc (strlen (s) + 1);
  if (p != NULL)
    (void) strcpy (p, s);

  return p;
}
#endif

void
fatal (char *format, ...)
{
  va_list ap;

  va_start (ap, format);
  fprintf (stderr, "%s: ", progname);
  vfprintf (stderr, format, ap);
  va_end (ap);
  fprintf (stderr, "\n");
  exit (EXIT_FAILURE);
}

void *
xmalloc (size_t size)
{
  void *p;

  p = malloc (size);
  if (p == NULL)
    fatal ("Memory allocate error");

  return p;
}

void *
xrealloc (void *p, size_t size)
{
  p = realloc (p, size);
  if (p == NULL)
    fatal ("Memory allocate error");

  return p;
}

char *
xstrdup (char *s)
{
  char *p;

  p = strdup (s);
  if (p == NULL)
    fatal ("Memory allocate error");

  return p;
}

FILE *
xfopen (char *fname, char *mode)
{
  FILE *fp;

  fp = fopen (fname, mode);
  if (fp == NULL)
    fatal ("Cannot open %s", fname);

  return fp;
}

#define TMPBUF_SIZE 256

char *
fgets_line (FILE *fp)
{
  char *dst, *p;
  int len, n;

  dst = p = xmalloc (TMPBUF_SIZE);
  dst[0] = '\0';
  len = 0;

  while (fgets (p, TMPBUF_SIZE, fp) != NULL)
    {
      n = strlen (p);
      len += n;
      if (n + 1 < TMPBUF_SIZE || dst[len - 1] == '\n')
        return xrealloc (dst, len + 1);
      else
        dst = xrealloc (dst, len + TMPBUF_SIZE);
      p = dst + len;
    }

  if (feof (fp) && len != 0)
    {
      dst = xrealloc (dst, len + 2);
      dst[len] = '\n';
      dst[len + 1] = '\0';
      return dst;
    }

  free (dst);
  return NULL;
}
