/*
 * Copyright (C) 1995, jack
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

static char rcsid[]= "$Id: main.c,v 1.17 1995/11/14 09:13:08 jack Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "getopt.h"
#include "makeargv.h"
#include "rsio.h"
#include "tilde.h"
#include "utils.h"
#include "version.h"
#include "zwork.h"

int verbose = 0;
char *progname = NULL;

#define DEFAULT_DIR "."
#define DEFAULT_LOG "/var/log/zlog"
#define DEFAULT_HOST "Unknown Host"
#define DEFAULT_WAIT (10*100)

static struct option long_options[] =
{
  {"directory", required_argument, NULL, 'd'},
  {"host", required_argument, NULL, 'H'},
  {"log", required_argument, NULL, 'l'},
  {"wait", required_argument, NULL, 'w'},
  {"crc", no_argument, NULL, 'c'},
  {"escape-control", no_argument, NULL, 'e'},
  {"hide-bar", no_argument, &Zhidebar, 1},
  {"no-log", no_argument, &Zlogtype, -1},
  {"no-resume", no_argument, &Zresume, 0},
  {"resume", no_argument, NULL, 'r'},
  {"quiet", no_argument, NULL, 'q'},
  {"short-crc", no_argument, NULL, 's'},
  {"tmn-type", no_argument, &Zlogtype, 1},
  {"verbose", no_argument, NULL, 'v'},
  {"help", no_argument, NULL, 'h'},
  {"version", no_argument, NULL, 'V'},
  {NULL, 0, NULL, 0}
};

static char *short_options = "ced:H:l:rw:qsvhV";

void version (void);
void usage (int send);
void help (int send);
int main (int argc, char *argv[]);

extern int zr (void);
extern int zs (char *files[]);

void
version (void)
{
  fprintf (stderr, "%s %s  Copyright (C) 1995, jack\n", progname, VERSION);
}

void
usage (int send)
{
  if (!send)
    {
      fprintf (stderr, "usage: %s [-cerqsvhV] [-d dir] [-H host] [-l log] [-w time]\n",
	       progname);
      fprintf (stderr, "       [--directory=dir] [--host=host] [--log=log] [--wait=time]\n");
      fprintf (stderr, "       [--crc] [--escape-control] [--hide-bar] [--no-log] [--no-resume]\n");
      fprintf (stderr, "       [--resume] [--quiet] [--short-crc] [--tmn-type] [--verbose]\n");
      fprintf (stderr, "       [--help] [--version]\n");
    }
  else
    {
      fprintf (stderr, "usage: %s [-cerqsvhV] [-H host] [-l log] [-w time] [--host=host]\n",
	       progname);
      fprintf (stderr, "       [--host=host] [--log=log] [--wait=time] [--crc]\n");
      fprintf (stderr, "       [--escape-control] [--hide-bar] [--no-log] [--no-resume]\n");
      fprintf (stderr, "       [--resume] [--quiet] [--short-crc] [--tmn-type] [--verbose]\n");
      fprintf (stderr, "       [--help] [--version] files...\n");
    }
}

void
help (int send)
{
  static char *help_message[]=
  {
    " -d dir --directory=dir down load directory (default: .)",
    " -H host --host=host    host name (default: $ZLOGHOST or \"Unknown Host\")",
    " -l log --log=log       log file (default: $ZLOG or /var/log/zlog)",
    " -w time --wait=time    wait time (default: 1000)",
    " -c --crc               crc check(in resume sequence)",
    " -e --escape-control    escape all control characters",
    " --hide-bar             don't show a bar of transferred size",
    " --no-log               don't output logfile",
    " --no-resume            don't resume",
    " -r --resume            resume request",
    " -q --quiet             quiet mode",
    " -s --short-crc         use 16bit CRC",
    " --tmn-type             use a form of filexfer.log",
    " -v --verbose           verbose mode",
    " -h --help              give this help",
    " -V --version           display version number",
    NULL
  };
  static char *send_help_message[]=
  {
    " -H host --host=host    host name (default: $ZLOGHOST or \"Unknown Host\")",
    " -l log --log=log       log file (default: $ZLOG or /var/log/zlog)",
    " -w time --wait=time    wait time (default: 1000)",
    " -c --crc               crc check(in resume sequence)",
    " -e --escape-control    escape all control characters",
    " --hide-bar             don't show a bar of transferred size",
    " --no-log               don't output logfile",
    " --no-resume            don't resume",
    " -r --resume            resume request",
    " -q --quiet             quiet mode",
    " -s --short-crc         use 16bit CRC",
    " --tmn-type             use a form of filexfer.log",
    " -v --verbose           verbose mode",
    " -h --help              give this help",
    " -V --version           display version number",
    NULL
  };
  char **p = send ? send_help_message : help_message ;

  version ();
  usage (send);
  while (*p)
    fprintf (stderr, "%s\n", *p++);
}

int
main (int argc, char *argv[])
{
  struct stat statbuf;
  char **envargv = NULL, **newargv, **tmpargv;
  char *p;
  int send = 0;
  int ch;
  int rc = 0;

  progname = xstrdup (get_basename (argv[0]));
#if defined (SUPPRESS_EXTENSION)
  (void) suppress_extension (progname);
#endif
  argv[0] = xstrdup (progname);

  if (strnicmp (progname, "zs", sizeof "zs" - 1) == 0)
    send = 1;

  if ((p = getenv ("ZOPT")) != NULL)
    envargv = makeargv (p);
  if (!send && (p = getenv ("ZROPT")) != NULL)
    {
      tmpargv = makeargv (p);
      newargv = addargv (envargv, tmpargv);
      freeargv (tmpargv);
      freeargv (envargv);
      envargv = newargv;
    }
  else if ((p = getenv ("ZSOPT")) != NULL)
    {
      tmpargv = makeargv (p);
      newargv = addargv (envargv, tmpargv);
      freeargv (tmpargv);
      freeargv (envargv);
      envargv = newargv;
    }
  argv = insertargv (argv, envargv);
  freeargv (envargv);
  argc = countargv (argv);

  Zrwait = DEFAULT_WAIT;
  Zrdir = NULL;
  Zlogtype = 0;
  Zlog = NULL;
  Zloghost = NULL;

  while ((ch = getopt_long (argc, argv, short_options, long_options, (int *) 0)) != EOF)
    {
      switch (ch)
	{
	case 0:
	  break;
	case 'd':
	  if (send)
	    {
	      usage (send);
	      exit (EXIT_FAILURE);
	    }
	  if (Zrdir != NULL)
	    free (Zrdir);
	  Zrdir = xstrdup (optarg);
	  toslash (Zrdir);
	  break;
	case 'H':
	  if (Zloghost != NULL)
	    free (Zloghost);
	  Zloghost = xstrdup (optarg);
	  break;
	case 'l':
	  if (Zlog != NULL)
	    free (Zlog);
	  Zlog = xstrdup (optarg);
	  toslash (Zlog);
	  if (Zlogtype < 0)
	    Zlogtype = 0;
	  break;
	case 'w':
	  Zrwait = atoi (optarg);
	  if (Zrwait < 1)
	    Zrwait = 1;
	  else if (Zrwait > 65535)
	    Zrwait = 65535;
	  break;
	case 'c':
	  Zcheckcrc = 1;
	  break;
	case 'e':
	  Zctlesc = 1;
	  break;
	case 'r':
	  Zresume = 1;
	  break;
	case 'q':
	  verbose = -1;
	  break;
	case 's':
	  Zshortcrc = 1;
	  break;
	case 'v':
	  verbose = 1;
	  break;
	case 'h':
	  help (send);
	  exit (EXIT_FAILURE);
	  break;
	case 'V':
	  version ();
	  exit (EXIT_SUCCESS);
	  break;
	default:
	  usage (send);
	  exit (EXIT_FAILURE);
	  break;
	}
    }

  if (send && optind >= argc)
    {
      usage (send);
      exit (EXIT_FAILURE);
    }

  if (Zrdir == NULL)
    Zrdir = DEFAULT_DIR;
  Zrdir = tilde_expand (Zrdir);
  if (!send)
    if (stat (Zrdir, &statbuf) != 0 || !(statbuf.st_mode & S_IFDIR))
      fatal ("%s: No such directory", Zrdir);

  if (Zlog == NULL)
    {
      if ((Zlog = getenv ("ZLOG")) != NULL)
	{
	  tmpargv = makeargv (Zlog);
	  if (tmpargv[0] != NULL)
	    Zlog = xstrdup (tmpargv[0]);
	  else
	    Zlog = DEFAULT_LOG;
	  freeargv (tmpargv);
	}
      else
	Zlog = DEFAULT_LOG;
    }
  Zlog = tilde_expand (Zlog);
  if (Zloghost == NULL)
    {
      if ((Zloghost = getenv ("ZLOGHOST")) != NULL)
	{
	  tmpargv = makeargv (Zloghost);
	  if (tmpargv[0] != NULL)
	    Zloghost = xstrdup (tmpargv[0]);
	  else
	    Zloghost = DEFAULT_HOST;
	  freeargv (tmpargv);
	}
      else
	Zloghost = DEFAULT_HOST;
    }
  if (Zlogtype >= 0)
    {
      if (verbose > 0)
	printf ("Logfile%s = \"%s\"\n", Zlogtype ? "(tmn)": "", Zlog);
      Zlogfp = xfopen (Zlog, FOPEN_APPENDBINARY_PLUS);
    }
  else
    Zlogfp = NULL;
  RSinWaitTime = Zrwait;

  if (RSinit () < 0)
    fatal ("Initialize error");

  if (!send)
    {
      if (verbose > 0)
	printf ("Down Load Directory --> \"%s\"\n", Zrdir);
      rc = zr ();
      rc &= 0x7f;
    }
  else
    {
      rc = zs (&argv[optind]);
      rc &= 0x7f;
    }

  if (Zlogfp != NULL)
    fclose (Zlogfp);
  RSend ();

  return rc;
}
