/*
 * Copyright (C) 1995, jack
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

static char rcsid[]= "$Id: rsio.c,v 1.8 1995/10/29 12:17:32 jack Exp $";

#include <stdlib.h>
#include <signal.h>
#include <sys/dos.h>
#include <sys/iocs.h>

#include "config.h"
#ifndef NO_RSIO_INLINE
#define NO_RSIO_INLINE
#endif
#include "rsio.h"
#include "keyin.h"
#include "zsend.h"

#define OK (0)
#define ERROR (-1)
#define TIMEOUT (-2)
#define ESC (0x1b)

static int RSssp = -1;
int RSinWaitTime = 1000;

int
RSinit (void)
{
  int ssp;

  if ((ssp = _dos_super (0)) >= 0)
    RSssp = ssp;
  atexit (RSend);

  if (_iocs_set232c (-1) & 0x0040)
    {
      (void) _iocs_set232c (0x0aff);
      (void) _iocs_set232c (0x0cff);
    }

  return OK;
}

void
RSend (void)
{
  if (_iocs_set232c (-1) & 0x0040)
    {
      (void) _iocs_set232c (0x0bff);
      (void) _iocs_set232c (0x0dff);
    }

  if (RSssp >= 0)
    {
      _dos_super (RSssp);
      RSssp = -1;
    }
}

int
RSKeyCheck (void)
{
  while (KeySense ())
    if (Keyin () == ESC)
      {
	raise (SIGTERM);
	return ERROR;
      }

  return OK;
}

int
RSin (void)
{
  int c;

  if (_iocs_lof232c () & 0x0000ffff)
    c = _iocs_inp232c () & 0x000000ff;
  else if (RSinWait () >= 0)
    c = _iocs_inp232c () & 0x000000ff;
  else
    c = TIMEOUT;

  return c;
}

int
RSinWait (void)
{
  int base, passed;

  base = _iocs_ontime ();

  do
    {
      if (_iocs_bitsns (0) & (1 << 1))
	if (RSKeyCheck () < 0)
	  return ERROR;

      if (_iocs_lof232c () & 0x0000ffff)
	return OK;

      if ((passed = _iocs_ontime () - base) < 0)
	{
	  passed += 8640000;
	  base -= 8640000;
	}
    }
  while (passed < RSinWaitTime);

  return TIMEOUT;
}

int
RSinSense (void)
{
  if (_iocs_lof232c () & 0x0000ffff)
    return 1;

  return 0;
}

int
RSout (int ch)
{
  if (_iocs_osns232c () != 0)
    _iocs_out232c (ch);
  else if (RSoutWait () >= 0)
    _iocs_out232c (ch);
  else
    return TIMEOUT;

  return OK;
}

int
RSoutWait (void)
{
  do
    {
      if (_iocs_bitsns (0) & (1 << 1))
	if (RSKeyCheck () < 0)
	  return ERROR;
    }
  while (_iocs_osns232c () == 0);

  return OK;
}

int
RSoutSense (void)
{
  if (_iocs_osns232c () != 0)
    return 1;

  return 0;
}

void
RSflush (void)
{
}

int
RSprint (char *s)
{
  int c;

  for (; *s; s++)
    if ((c = RSout (*s)) < 0)
      return c;

  return OK;
}
