/*
 * Copyright (C) 1995, jack
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

static char rcsid[]= "$Id: zmisc.c,v 1.4 1995/11/05 21:04:52 jack Exp $";

#include <stdio.h>
#include <time.h>

#include "config.h"
#include "utils.h"
#include "zmisc.h"
#include "zmodem.h"
#include "zreceive.h"
#include "zwork.h"

#define MAXBARS (44-1)

void
zvhprint (struct zwork *zworkp, int type, char *format)
{
  if (format[0] == 'S' || format[0] == 's')
    {
      printf (format, zgettype(type), zworkp->sbuf[0] & 0xff,
	      zworkp->sbuf[1] & 0xff, zworkp->sbuf[2] & 0xff,
	      zworkp->sbuf[3] & 0xff);
    }
  else
    {
      printf (format, zgettype(type), zworkp->rbuf[0] & 0xff,
	      zworkp->rbuf[1] & 0xff, zworkp->rbuf[2] & 0xff,
	      zworkp->rbuf[3] & 0xff);
    }
}

char *
zgettype (int type)
{
  static char *typestr[] =
    { "ZRQINIT", "ZRINIT", "ZSINIT", "ZACK", "ZFILE", "ZSKIP", "ZNAK",
      "ZABORT", "ZFIN", "ZRPOS", "ZDATA", "ZEOF", "ZFERR", "ZCRC",
      "ZCHALLENGE", "ZCOMPL", "ZCAN", "ZFREECNT", "ZCOMMAND", "ZSTDERR",
    };
  static char *typestre[] = { "NO TYPE", "ERROR", "TIMEOUT", "RCDO", "GCOUNT" };

  if (type < ZRQINIT || type > ZSTDERR)
    {
      if (type < GCOUNT || type > ERROR)
	return typestre[0];
      else
	return typestre[-type];
    }

  return typestr[type];
}

void
zbarprint (int fsize, int rsize, int tssize, clock_t ttotal)
{
  /*             12345678901234567890123456789012345678901234567890 */
  char *tbarl = "===========================================";
  char *tbars = "                                           ";
  static int oldshow = -1;
  int per;
  int bars;
  int rest;
  double rate;

  if (!Zhidebar)
    {
      if (fsize >= 500*1024)
	{
	  per = fsize ? (rsize * 1000 / fsize) : 1000;
	  if (oldshow == per)
	    return;
	  oldshow = per;
	  printf ("%3d.%1.1d%% ", per / 10, per % 10);
	  bars = (MAXBARS * per) / 1000;
	}
      else
	{
	  per = fsize ? (rsize * 100 / fsize) : 100;
	  if (oldshow == per)
	    return;
	  oldshow = per;
	  printf ("%3d%% ", per);
	  bars = (MAXBARS * per) / 100;
	}
      printf ("%.*s%c%.*s %d bytes ", bars, tbarl, per ? '>' : ' ',
	      (MAXBARS - bars), tbars, fsize);
    }
  else
    {
      if (oldshow == rsize)
	return;
      oldshow = rsize;
      printf ("%d/%d bytes ", rsize, fsize);
    }

  rate = ttotal ? ((double) ((rsize - tssize) * CLK_TCK) / (double) ttotal) : 0.0;
  printf ("%8.2f bytes/s ", rate);
  if (rate)
    {
      rest = (fsize - rsize) / rate;
      printf ("R:%3d:%2.2d", rest / 60, rest % 60);
    }
  else
    printf ("R: --:--");
  putchar (isatty (STDOUT_FILENO) ? '\r' : '\n');
  fflush (stdout);
}

void
zlogout (int mode, char *fname, time_t tcomp, int tsize, clock_t ttime)
{
  char tbuf[128];
  int tsec;

  switch (Zlogtype)
    {
    case 0:
      strftime (tbuf, sizeof tbuf, "%b %d %Y %H:%M", localtime (&tcomp));
      fprintf (Zlogfp, "%s %s: %s %d bytes %.2f sec. %.2f bytes/s\n",
	       tbuf, (mode == 0) ? "Recv" : "Send", fname, tsize,
	       ((double) ttime / (double) CLK_TCK),
	       ((double) (tsize * CLK_TCK) / (double) ttime));
      break;
    case 1:
      tsec = (int) (ttime / CLK_TCK);
      strftime (tbuf, sizeof tbuf, "%y%m%d %H%M%S", localtime (&tcomp));
      fprintf (Zlogfp, "%s: [%-20.20s][%-22.22s] %s %s %7dsec(%2d:%2.2d:%2.2d) %8dbyte %5db/s\r\n",
	       tbuf, Zloghost, get_basename (fname), "ZM ", (mode == 0) ? "DW" : "UP",
	       tsec, (tsec / 3600), ((tsec / 60) % 60), (tsec % 60),
	       tsize,
	       ((int) (tsize * CLK_TCK) / (int) ttime));
      break;
    default:
      break;
    }
}
