/*
 * Copyright (C) 1995, jack
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

static char rcsid[]= "$Id: zr.c,v 1.11 1995/11/14 09:13:47 jack Exp $";

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <utime.h>

#include "config.h"
#include "keyin.h"
#include "rsio.h"
#include "updcrc.h"
#include "utils.h"
#include "zsend.h"
#include "zreceive.h"
#include "zmisc.h"
#include "zmodem.h"
#include "zwork.h"

extern int verbose;

#define ITOB(a,x) \
do \
  { \
    int xx = (x); \
    (a)[ZP0] = ((xx) & 0xff); \
    (a)[ZP1] = ((xx >>= 8) & 0xff); \
    (a)[ZP2] = ((xx >>= 8) & 0xff); \
    (a)[ZP3] = ((xx >>= 8) & 0xff); \
  } \
while (0)
#define VNNEWLINE() \
do \
  if (verbose >= 0 && isatty (STDOUT_FILENO)) \
    printf ("\n"); \
while (0)
#define UPDATEBAR(fsize,rsize,tssize,ttotal) \
do \
  if (verbose >= 0) \
    zbarprint ((fsize), (rsize), (tssize), (ttotal)); \
while (0)
#define VHPRINT(workp,type,format) \
do \
  if (verbose > 0) \
    zvhprint ((workp), (type), (format)); \
while (0)
#define VPRINTF(format) \
do \
  if (verbose > 0) \
    printf format; \
while (0)
#define VNPRINTF(format) \
do \
  if (verbose >= 0) \
    printf format; \
while (0)
#define Seek(fd,off,mod) lseek (fd, off, mod)
#define Read(fd,buf,len) read (fd, buf, len)
#define Write(fd,buf,len) write (fd, buf, len)
#define Close(fd) close (fd)
#define get_filebuf() (writebuf)
#define get_filebufsiz() (writebufsize)

#define OK (0)
#define ESC (0x1b)
#define BUFFERED_SIZE (16384)

static void quit (int sig);
int zr (void);
int zrtry (int hdrtype);
void zrsendfin (void);
#if 0
int zspawn (char *str);
int zexec (char *str);
#endif
int zreceivefiles (void);
int zreceivefile (void);
int zprocheader (char *info);
long zstrtol (char *str, char **end, int base);

static void (*Original_sigint) (int) = NULL;
static void (*Original_sigquit) (int) = NULL;
static void (*Original_sigterm) (int) = NULL;
static char *writebuf = NULL;
static int writebufsize = BUFFERED_SIZE;

static void
quit (int sig)
{
  struct zwork *zworkp = &zwork;

  zcancel ();
  RSend ();
  Keyflush ();
  if (zworkp->fd != -1 && zworkp->bpos != 0)
    {
      Write (zworkp->fd, zworkp->buffer, zworkp->bpos);
      zworkp->bpos = 0;
    }
  if (writebuf != NULL)
    {
      free (writebuf);
      writebuf = NULL;
    }
  exit (1);
}

int
zr (void)
{
  struct zwork *zworkp = &zwork;
  int c;

  zworkp->usefcs32 = 0;
  zworkp->ctlesc = Zctlesc;
  zworkp->retry = 0;
  zworkp->attn[0] = '\0';
  zworkp->fname[0] = '\0';
  zworkp->fd = -1;
  zworkp->fpos = 0;
  zworkp->bpos = 0;

  Original_sigint = signal (SIGINT, quit);
#ifdef SIGQUIT
  Original_sigquit = signal (SIGQUIT, quit);
#endif
  Original_sigterm = signal (SIGTERM, quit);
  writebuf = xmalloc (writebufsize);

  if ((c = zrtry (ZRINIT)) != 0)
    {
      if (c == ZCOMPL)
	goto exit_ok;
      if (c == ERROR)
	{
cancel:
	  if (zworkp->fd != -1)
	    {
	      struct utimbuf utimbuf;

	      Close (zworkp->fd);
	      zworkp->fd = -1;

	      if (zworkp->ftime)
		{
		  utimbuf.actime = time (NULL);
		  utimbuf.modtime = zworkp->ftime;
		  utime (zworkp->fname, &utimbuf);
		}
	    }
	  zcancel ();

	  (void) signal (SIGINT, Original_sigint);
#ifdef SIGQUIT
	  (void) signal (SIGQUIT, Original_sigquit);
#endif
	  (void) signal (SIGTERM, Original_sigterm);
	  free (writebuf);
	  writebuf = NULL;

	  return ERROR;
	}
    }

  if (zreceivefiles ())
    goto cancel;

exit_ok:
  (void) signal (SIGINT, Original_sigint);
#ifdef SIGQUIT
  (void) signal (SIGQUIT, Original_sigquit);
#endif
  (void) signal (SIGTERM, Original_sigterm);
  free (writebuf);
  writebuf = NULL;

  return OK;
}

int
zrtry (int hdrtype)
{
  struct zwork *zworkp = &zwork;
  int c;
  short n;

  for (n = 15; --n >= 0; )
    {
      ITOB (zworkp->sbuf, 0);
      if (hdrtype == ZRINIT)
	{
	  if (!Zshortcrc)
	    zworkp->sbuf[ZF0] = CANFC32 | CANFDX | CANOVIO;
	  else
	    zworkp->sbuf[ZF0] = CANFDX | CANOVIO;
	  if (zworkp->ctlesc)
	    zworkp->sbuf[ZF0] |= ESCCTL;
	  VHPRINT (zworkp, hdrtype, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
	}
      else
	{
	  VHPRINT (zworkp, hdrtype, "Send: %s\n");
	}
      zshhdr (4, hdrtype, zworkp->sbuf);
      if (hdrtype == ZSKIP)
	hdrtype = ZRINIT;
again:
      if (KeySense ())
	if (Keyin () == ESC)
	  raise (SIGTERM);

      switch ((c = zgethdr (zworkp->rbuf)))
	{
	case ZRQINIT:
	  VHPRINT (zworkp, ZRQINIT, "Recv: %s\n");
	  continue;
	case ZEOF:
	  VHPRINT (zworkp, ZEOF, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  continue;
	case TIMEOUT:
	  VHPRINT (zworkp, TIMEOUT, "Recv: %s\n");
	  continue;
	case ZFILE:
	  VHPRINT (zworkp, ZFILE, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  zworkp->zconv = zworkp->rbuf[ZF0];
	  zworkp->zmanag = zworkp->rbuf[ZF1];
	  zworkp->ztrans = zworkp->rbuf[ZF2];
	  hdrtype = ZRINIT;
	  if (zrdata (zworkp->rdbuf, ZMAXSPLEN) == GOTCRCW)
	    return ZFILE;
	  ITOB (zworkp->sbuf, 0);
	  VHPRINT (zworkp, ZNAK, "Send: %s\n");
	  zshhdr (4, ZNAK, zworkp->sbuf);
	  goto again;
	case ZSINIT:
	  VHPRINT (zworkp, ZSINIT, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  zworkp->ctlesc |= zworkp->rbuf[ZF0] & TESCCTL;
	  if (zrdata (zworkp->attn, ZATTNLEN) == GOTCRCW)
	    {
	      ITOB (zworkp->sbuf, 1);
	      VHPRINT (zworkp, ZACK, "Send: %s\n");
	      zshhdr (4, ZACK, zworkp->sbuf);
	      goto again;
	    }
	  ITOB (zworkp->sbuf, 0);
	  VHPRINT (zworkp, ZNAK, "Send: %s\n");
	  zshhdr (4, ZNAK, zworkp->sbuf);
	  goto again;
	case ZFREECNT:
	  VHPRINT (zworkp, ZFREECNT, "Recv: %s\n");
	  ITOB (zworkp->sbuf, 0);
	  VHPRINT (zworkp, ZACK, "Send: %s\n");
	  zshhdr (4, ZACK, zworkp->sbuf);
	  goto again;
	case ZCOMMAND:
	  {
	    short flag;

	    VHPRINT (zworkp, ZCOMMAND, "Recv: %s\n");
	    flag = zworkp->rbuf[ZF0];
	    if (zrdata (zworkp->rdbuf, ZMAXSPLEN) == GOTCRCW)
	      {
#if 0
		if (flag & ZCACK1)
		  ITOB (zworkp->sbuf, 0);
		else
		  ITOB (zworkp->sbuf, zspawn (zworkp->rdbuf));
#else
		ITOB (zworkp->sbuf, 0);
#endif
		do
		  {
		    VHPRINT (zworkp, ZCOMPL, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
		    zshhdr (4, ZCOMPL, zworkp->sbuf);
		  }
		while (++zworkp->retry < 20 && zgethdr (zworkp->rbuf) != ZFIN);
		zrsendfin ();
#if 0
		if (flag & ZCACK1)
		  zexec (zworkp->rdbuf);
#endif
		return ZCOMPL;
	      }
	    ITOB (zworkp->sbuf, 0);
	    VHPRINT (zworkp, ZNAK, "Send: %s\n");
	    zshhdr (4, ZNAK, zworkp->sbuf);
	    goto again;
	  }
	case ZCOMPL:
	  VHPRINT (zworkp, ZCOMPL, "Recv: %s\n");
	  goto again;
	default:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  continue;
	case ZFIN:
	  VHPRINT (zworkp, ZFIN, "Recv: %s\n");
	  zrsendfin ();
	  return ZCOMPL;
	case ZCAN:
	  VHPRINT (zworkp, ZCAN, "Recv: %s\n");
	  return ERROR;
	}
    }

  return OK;
}

void
zrsendfin (void)
{
  struct zwork *zworkp = &zwork;
  short n;

  ITOB (zworkp->sbuf, 0);
  for (n = 3; --n >= 0; )
    {
      VHPRINT (zworkp, ZFIN, "Send: %s\n");
      zshhdr (4, ZFIN, zworkp->sbuf);
      switch (RSin ())
	{
	case 'O':
	  RSin ();
	  return;
	case RCDO:
	  return;
	case TIMEOUT:
	default:
	  break;
	}
    }
}

#if 0
int
zspawn (char *str)
{
  if (*str == '!')
    str++;
  return system (str);
}

int
zexec (char *str)
{
  if (*str == '!')
    str++;
  return execl ("/bin/sh", "sh", "-c", str);
}
#endif

int
zreceivefiles (void)
{
  struct zwork *zworkp = &zwork;
  int c;

  while (1)
    {
      switch (c = zreceivefile ())
	{
	case ZEOF:
	  c = ZRINIT;
	case ZSKIP:
	  switch (zrtry (c))
	    {
	    case ZCOMPL:
	      return OK;
	    default:
	      return ERROR;
	    case ZFILE:
	      zworkp->retry = 0;
	      zworkp->fname[0] = '\0';
	      zworkp->fd = -1;
	      zworkp->fpos = 0;
	      zworkp->bpos = 0;
	      break;
	    }
	  continue;
	default:
	  return c;
	case ERROR:
	  return ERROR;
	}
    }
}

int
zreceivefile (void)
{
  struct zwork *zworkp = &zwork;
  clock_t tstime, ttime, tdstime, tdtime, tdtotal;
  time_t tcomp;
  int tssize, tsize;
  int c;
  int rsize;
  short maxerrs;

  if ((c = zprocheader (zworkp->rdbuf)) != 0)
    return c;

  zworkp->bcount = 0;
  zworkp->bpos = 0;
  zworkp->bufsiz = get_filebufsiz ();
  zworkp->buffer = get_filebuf ();

  maxerrs = 20;
  rsize = zworkp->fpos;
  tssize = rsize;
  tdtime = 0;
  tdtotal = 0;
  tstime = clock ();

  VNPRINTF (("Receiving file: %s\n", zworkp->fname));
  while (1)
    {
      ITOB (zworkp->sbuf, rsize);
      VHPRINT (zworkp, ZRPOS, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
      zshhdr (4, ZRPOS, zworkp->sbuf);
nextheader:
      if (KeySense ())
	if (Keyin () == ESC)
	  raise (SIGTERM);

      switch (c = zgethdr (zworkp->rbuf))
	{
	default:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  if (--maxerrs < 0)
	    {
	      UPDATEBAR (zworkp->fsize, rsize, tssize, tdtotal);
	      VNNEWLINE ();
	      return ERROR;
	    }
	  continue;
	case ZCAN:
	  VHPRINT (zworkp, ZCAN, "Recv: %s\n");
	  return ERROR;
	case ZNAK:
	  VHPRINT (zworkp, ZNAK, "Recv: %s\n");
	  if (--maxerrs < 0)
	    {
	      UPDATEBAR (zworkp->fsize, rsize, tssize, tdtotal);
	      VNNEWLINE ();
	      return ERROR;
	    }
	  continue;
	case TIMEOUT:
	  VHPRINT (zworkp, TIMEOUT, "Recv: %s\n");
	  if (--maxerrs < 0)
	    {
	      UPDATEBAR (zworkp->fsize, rsize, tssize, tdtotal);
	      VNNEWLINE ();
	      return ERROR;
	    }
	  continue;
	case ZFILE:
	  VHPRINT (zworkp, ZFILE, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  zrdata (zworkp->rdbuf, ZMAXSPLEN);
	  continue;
	case ZEOF:
	  UPDATEBAR (zworkp->fsize, rsize, tssize, tdtotal);
	  VNNEWLINE ();
	  VHPRINT (zworkp, ZEOF, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  if (zworkp->rpos != rsize)
	    {
	      zworkp->retry = 0;
	      goto nextheader;
	    }
	  if (zworkp->fd != -1)
	    {
	      struct utimbuf utimbuf;

	      Close (zworkp->fd);
	      zworkp->fd = -1;

	      if (zworkp->ftime)
		{
		  utimbuf.actime = time (NULL);
		  utimbuf.modtime = zworkp->ftime;
		  utime (zworkp->fname, &utimbuf);
		}
	    }

	  ttime = clock () - tstime;
	  tcomp = time (NULL);
	  if (ttime == 0)
	    ttime = 1;
	  tsize = rsize - tssize;
	  VNPRINTF (("%s: %d bytes %.2f sec. %.2f bytes/s\n", zworkp->fname, tsize,
		    ((double) ttime / (double) CLK_TCK),
		    ((double) (tsize * CLK_TCK) / (double) ttime)));
	  zlogout (0, zworkp->fname, tcomp, tsize, ttime);
	  return ZEOF;
	case ERROR:
	  VHPRINT (zworkp, ERROR, "Recv: %s\n");
	  if (--maxerrs < 0)
	    {
	      UPDATEBAR (zworkp->fsize, rsize, tssize, tdtotal);
	      VNNEWLINE ();
	      return ERROR;
	    }
	  RSprint (zworkp->attn);
	  continue;
	case ZSKIP:
	  VHPRINT (zworkp, ZSKIP, "Recv: %s\n");
	  return ZSKIP;
	case ZDATA:
	  VHPRINT (zworkp, ZDATA, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  if (zworkp->rpos != rsize)
	    {
	      if (--maxerrs < 0)
		return ERROR;
	      RSprint (zworkp->attn);
	      continue;
	    }
	  tdstime = clock ();
moredata:
	  if (KeySense ())
	    if (Keyin () == ESC)
	      raise (SIGTERM);

	  UPDATEBAR (zworkp->fsize, rsize, tssize, tdtotal + tdtime);
	  c = zrdata (zworkp->buffer + zworkp->bpos, ZMAXSPLEN);
	  tdtime = clock () - tdstime;
	  switch (c)
	    {
	    case ZCAN:
	      VHPRINT (zworkp, ZCAN, "Recv: %s\n");
	      Write (zworkp->fd, zworkp->buffer, zworkp->bpos);
	      zworkp->bpos = 0;
	      tdtotal += tdtime;
	      UPDATEBAR (zworkp->fsize, rsize, tssize, tdtotal);
	      VNNEWLINE ();
	      return ERROR;
	    case ERROR:
	      VHPRINT (zworkp, ERROR, "Recv: %s\n");
	      Write (zworkp->fd, zworkp->buffer, zworkp->bpos);
	      zworkp->bpos = 0;
	      tdtotal += tdtime;
	      if (--maxerrs < 0)
		{
		  UPDATEBAR (zworkp->fsize, rsize, tssize, tdtotal);
		  VNNEWLINE ();
		  return ERROR;
		}
	      RSprint (zworkp->attn);
	      continue;
	    case TIMEOUT:
	      VHPRINT (zworkp, TIMEOUT, "Recv: %s\n");
	      Write (zworkp->fd, zworkp->buffer, zworkp->bpos);
	      zworkp->bpos = 0;
	      tdtotal += tdtime;
	      if (--maxerrs < 0)
		{
		  UPDATEBAR (zworkp->fsize, rsize, tssize, tdtotal);
		  VNNEWLINE ();
		  return ERROR;
		}
	      continue;
	    default:
	      VHPRINT (zworkp, c, "Recv: %s\n");
	      Write (zworkp->fd, zworkp->buffer, zworkp->bpos);
	      zworkp->bpos = 0;
	      tdtotal += tdtime;
	      break;
	    case GOTCRCW:
	      maxerrs = 20;
	      zworkp->bpos += zworkp->rdcount;
	      Write (zworkp->fd, zworkp->buffer, zworkp->bpos);
	      zworkp->bpos = 0;
	      rsize += zworkp->rdcount;
	      tdtotal += tdtime;
	      ITOB (zworkp->sbuf, rsize);
	      VHPRINT (zworkp, ZACK, "Send: %s\n");
	      zshhdr (4, ZACK, zworkp->sbuf);
	      RSout (XON);
	      goto nextheader;
	    case GOTCRCQ:
	      maxerrs = 20;
	      zworkp->bpos += zworkp->rdcount;
	      if ((zworkp->bpos + ZMAXSPLEN) > zworkp->bufsiz)
		{
		  Write (zworkp->fd, zworkp->buffer, zworkp->bpos);
		  zworkp->bpos = 0;
		}
	      rsize += zworkp->rdcount;
	      ITOB (zworkp->sbuf, rsize);
	      VHPRINT (zworkp, ZACK, "Send: %s\n");
	      zshhdr (4, ZACK, zworkp->sbuf);
	      goto moredata;
	    case GOTCRCG:
	      maxerrs = 20;
	      zworkp->bpos += zworkp->rdcount;
	      if ((zworkp->bpos + ZMAXSPLEN) > zworkp->bufsiz)
		{
		  Write (zworkp->fd, zworkp->buffer, zworkp->bpos);
		  zworkp->bpos = 0;
		}
	      rsize += zworkp->rdcount;
	      goto moredata;
	    case GOTCRCE:
	      maxerrs = 20;
	      zworkp->bpos += zworkp->rdcount;
	      Write (zworkp->fd, zworkp->buffer, zworkp->bpos);
	      zworkp->bpos = 0;
	      rsize += zworkp->rdcount;
	      tdtotal += tdtime;
	      goto nextheader;
	    }
	}
    }
}

int
zprocheader (char *info)
{
  struct zwork *zworkp = &zwork;
  struct stat statbuf;
  char newbase[FILENAME_MAX];
  char newname[FILENAME_MAX];
  char *p, *q;
  short n;
  int fd;
  long value, Filesize = 0, Modtime = 0;

  zworkp->fsize = 0;
  zworkp->ftime = 0;

  strcpy (zworkp->fname, Zrdir);
  p = zworkp->fname + strlen (Zrdir);
  if (p != zworkp->fname && p[-1] != '/')
    *p++ = '/';
  q = info;
  if (q[0] == '/')
    q++;
  for (n = 63; --n >= 0; )
    *p++ = *q++;
  *p++ = '\0';

  p = info + strlen (info) + 1;
  value = zstrtol (p, &q, 10);
  if (p != q)
    {
      Filesize = value;
      p = q;
      value = zstrtol (p, &q, 8);
      if (p != q)
	Modtime = value;
    }

  zworkp->fsize = Filesize;
  zworkp->ftime = Modtime;

  if (stat (zworkp->fname, &statbuf) != 0)
    {
      if (zworkp->zmanag & ZMSKNOLOC)
	{
    skip:
	  VPRINTF (("Skip: %s\n", zworkp->fname));
	  return ZSKIP;
	}
      else
        goto newopen;
    renopen:
      strcpy (newbase, zworkp->fname);
      p = newbase + strlen (newbase);
      if ((p - newbase) > 3)
	p[-3] = '\0';
      for (n = 0; n < 1000; n++)
	{
	  sprintf (newname, "%s%3.3d", newbase, (int)n);
	  if (stat (newname, &statbuf) != 0)
	    break;
	}
      if (n >= 1000)
	return ZFERR;
      VPRINTF (("Rename: %s -> %s\n", zworkp->fname, newname));
      if (rename (zworkp->fname, newname) < 0)
	return ZFERR;
    newopen:
      VPRINTF (("Open: %s (new)\n", zworkp->fname));
      if ((fd = open (zworkp->fname, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,
		      S_IREAD|S_IWRITE)) < 0)
	{
	  p = zworkp->fname + strlen(zworkp->fname);
	  if ((p - zworkp->fname) > 1)
	    {
	      p[-1] = '\0';
	      if (stat (zworkp->fname, &statbuf) != 0)
		goto newopen;
	      goto renopen;
	    }
	  return ZFERR;
	}
    }
  else if (statbuf.st_mode & S_IFREG)
    {
      zworkp->zmanag &= ZMMASK;
      if ((zworkp->zconv == ZCRESUM) || Zresume)
	{
	  if (Filesize < statbuf.st_size)
	    goto renopen;
	  VPRINTF (("Open: %s (resume)\n", zworkp->fname));
	  if ((fd = open (zworkp->fname, O_RDWR|O_BINARY)) < 0)
	    return ZFERR;

	  if (Zresume && Zcheckcrc)
	    {
	      int rcount;
	      int c;
	      int bufsiz = get_filebufsiz ();
	      char *buffer = get_filebuf ();
	      unsigned long crc;

	      crc = 0xffffffffL;
	      Seek (fd, 0, SEEK_SET);
	      while ((rcount = Read (fd, buffer, bufsiz)) > 0)
		crc = mkcrc32 (buffer, rcount, crc);
	      crc = ~crc;

	      ITOB (zworkp->sbuf, statbuf.st_size + 1);
	      VHPRINT (zworkp, ZCRC, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
	      zshhdr (4, ZCRC, zworkp->sbuf);
	      switch (c = zgethdr (zworkp->rbuf))
		{
		case ZCRC:
		  VHPRINT (zworkp, ZCRC, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
		  if ((unsigned long) zworkp->rpos == crc)
		    break;
		  Close (fd);
		  goto renopen;
		default:
		  VHPRINT (zworkp, c, "Recv: %s\n");
		  Close (fd);
		  goto renopen;
		}
	    }

	  zworkp->fpos = statbuf.st_size;
	  if (Seek (fd, zworkp->fpos, SEEK_SET) < 0)
	    {
	      Close (fd);
	      return ZFERR;
	    }
	  VPRINTF (("Recovery: at %d\n", zworkp->fpos));
	}
      else if ((zworkp->zmanag == ZMNEW) || ((zworkp->zmanag == ZMNEWL) &&
					     Filesize <= statbuf.st_size))
	{
	  if (statbuf.st_mtime > Modtime)
	    goto skip;
	  goto renopen;
	}
      else
	{
	  switch (zworkp->zmanag & ZMMASK)
	    {
	    case ZMCLOB:
	      goto newopen;
	    case ZMAPND:
	      VPRINTF (("Open: %s (append)\n", zworkp->fname));
	      if ((fd = open (zworkp->fname, O_WRONLY|O_APPEND|O_BINARY)) < 0)
		return ZFERR;
	      if (Seek (fd, 0, SEEK_END) < 0)
		{
		  Close (fd);
		  return ZFERR;
		}
	      break;
	    case 0:
	      goto renopen;
	    default:
	      goto skip;
	    }
	}
    }
  else
    return ZFERR;

  zworkp->fd = fd;

  return 0;
}

long
zstrtol (char *str, char **end, int base)
{
  char *p, *save;
  short sign;
  int c;
  long value;

  if (base < 2 || base > 36)
    return 0;

  p = str;
  while (*p == ' ' || *p == '\t')
    p++;

  sign = 0;
  switch (*p)
    {
    case '-':
      sign = 1;
    case '+':
      p++;
      break;
    }

  value = 0;
  save = p;
  for (c = *p; c != '\0'; c = *++p)
    {
      if (c >= '0' && c <= '9')
	c -= '0';
      else if (c >= 'a' && c <= 'z')
	c -= 'a' - 10;
      else if (c >= 'A' && c <= 'Z')
	c -= 'A' - 10;
      else
	break;
      if (c >= base)
	break;

      value *= base;
      value += c;
    }

  if (end)
    *end = (p == save) ? str : p;

  return (sign) ? -value : value;
}
