/*
 * Copyright (C) 1995, jack
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

static char rcsid[]= "$Id: zreceive.c,v 1.3 1995/11/01 15:10:32 jack Exp $";

#include "config.h"
#include "rsio.h"
#include "updcrc.h"
#include "zreceive.h"
#include "zmodem.h"
#include "zwork.h"

#if defined (USE_ASM_ZDECODE)
extern unsigned short zedmkcrc (int *count, int *rc, int ctlesc, char *buffer,
				int len, unsigned short crc);
extern unsigned long zedmkcrc32 (int *count, int *rc, int ctlesc, char *buffer,
				 int len, unsigned long crc);
#endif

int
xeRSin (void)
{
  struct zwork *zworkp = &zwork;
  int c;

  while (1)
    {
      if ((c = RSin ()) & 0x60)
	return c;
      switch (c)
	{
	case ZDLE:
	  break;
	case XOFF:
	case XOFF | 0x80:
	case XON:
	case XON | 0x80:
	  continue;
	default:
	  if (zworkp->ctlesc && !(c & 0x60))
	    continue;
	  return c;
	}
      break;
    }

  while (1)
    {
      if ((c = RSin ()) < 0)
	return c;
      if (c == CAN && (c = RSin ()) < 0)
	return c;
      if (c == CAN && (c = RSin ()) < 0)
	return c;
      if (c == CAN && (c = RSin ()) < 0)
	return c;
      switch (c)
	{
	case CAN:
	  return GOTCAN;
	case ZCRCE:
	case ZCRCG:
	case ZCRCQ:
	case ZCRCW:
	  return (c | GOTOR);
	case ZRUB0:
	  return 0x7f;
	case ZRUB1:
	  return 0xff;
	case XOFF:
	case XOFF | 0x80:
	case XON:
	case XON | 0x80:
	  continue;
	default:
	  if (zworkp->ctlesc && ! (c & 0x60))
	    continue;
	  if ((c & 0x60) == 0x40)
	    return (c ^ 0x40);
	  break;
	}
      return ERROR;
    }
}

int
x7RSin (void)
{
  struct zwork *zworkp = &zwork;
  int c;

  while (1)
    {
      if ((c = RSin ()) < 0)
	return c;
      switch (c &= 0x7f)
	{
	case XON:
	case XOFF:
	  continue;
	default:
	  if (zworkp->ctlesc && !(c & (ZDLE|0x40)))
	    continue;
	case CR:
	case LF:
	case ZDLE:
	  return c;
	}
    }
}

int
xhRSin (void)
{
  int c, n;

  if ((c = x7RSin ()) < 0)
    return c;
  n = c - '0';
  if (n > 9)
    n -= ('a' - ':');
  if (n & ~0xf)
    return ERROR;

  if ((c = x7RSin ()) < 0)
    return c;
  c -= '0';
  if (c > 9)
    c -= ('a' - ':');
  if (c & ~0xf)
    return ERROR;

  c += (n << 4);
  return c;
}

int
zgethdr (char *hdr)
{
  struct zwork *zworkp = &zwork;
  int c;
  short n = 16384;
  short cancount;

  cancount = 5;
  while (1)
    {
      c = RSin ();
      switch (c)
	{
	case XON:
	case XON | 0x80:
	  continue;
	case TIMEOUT:
	  goto got_character;
	case CAN:
got_can:
	  if (--cancount <= 0)
	    {
	      c = ZCAN;
	      goto got_character;
	    }
	  switch (c = RSin ())
	    {
	    case TIMEOUT:
	      continue;
	    case ZCRCW:
	      switch (RSin ())
		{
		case TIMEOUT:
		  c = ERROR;
		  goto got_character;
		default:
		  goto again;
		}
	    case CAN:
	      if (--cancount <= 0)
		{
		  c = ZCAN;
		  goto got_character;
		}
	      continue;
	    default:
	      break;
	    }
	default:
again:
	  if (--n == 0)
	    {
	      c = GCOUNT;
	      goto got_character;
	    }
	  cancount = 5;
	  continue;
	case ZPAD:
	case ZPAD | 0x80:
	  /* header request "ZPAD ZPAD ZDLE ..." */
	  cancount = 5;
	  goto splat;
	}
    }
splat:
  switch (c = x7RSin ())
    {
    case ZPAD:
      goto splat;
    case TIMEOUT:
      goto got_character;
    default:
      goto again;
    case ZDLE:
      break;
    }

  zworkp->rtype = c = x7RSin ();
  switch (c)
    {
    case ZBIN32:
      zworkp->rcrc32 = 1; c = zrbhd32 (hdr); break;
    case ZBIN:
      zworkp->rcrc32 = 0; c = zrbhdr (hdr); break;
    case ZHEX:
      zworkp->rcrc32 = 0; c = zrhhdr (hdr); break;
    case RCDO:
    case TIMEOUT:
      goto got_character;
    case CAN:
      goto got_can;
    default:
      goto again;
    }

  {
    int pos;
    pos = (hdr[ZP3] & 0xff);
    pos = (pos << 8) + (hdr[ZP2] & 0xff);
    pos = (pos << 8) + (hdr[ZP1] & 0xff);
    pos = (pos << 8) + (hdr[ZP0] & 0xff);
    zworkp->rpos = pos;
  }

got_character:
  switch (c)
    {
    case GOTCAN:
      c = ZCAN;
    case ZNAK:
    case ZCAN:
    case ERROR:
    case TIMEOUT:
    case GCOUNT:
      break;
    }

  return c;
}

int
zrbhd32 (char *hdr)
{
  struct zwork *zworkp = &zwork;
  int c;
  short n;
  unsigned long crc;

  if ((c = xeRSin ()) & ~0xff)
    return c;
  zworkp->rtype = c;
  crc = updcrc32 (c, 0xffffffffL);

  /* read flag and CRC */
  for (n = 4 + 4; --n >= 0; )
    {
      if ((c = xeRSin ()) & ~0xff)
	return c;
      *hdr++ = c;
    }

  crc = mkcrc32 (hdr - 8, 8, crc);
  if (crc != 0xdebb20e3)
    return ERROR;

  return zworkp->rtype;
}

int
zrbhdr (char *hdr)
{
  struct zwork *zworkp = &zwork;
  int c;
  short n;
  unsigned short crc;

  if ((c = xeRSin ()) & ~0xff)
    return c;
  zworkp->rtype = c;
  crc = updcrc (c, 0);

  /* read flag and CRC */
  for (n = 4 + 2; --n >= 0; )
    {
      if ((c = xeRSin ()) & ~0xff)
	return c;
      *hdr++ = c;
    }

  crc = mkcrc (hdr - 6, 6, crc);
  if (crc & 0xffff)
    return ERROR;

  return zworkp->rtype;
}

int
zrhhdr (char *hdr)
{
  struct zwork *zworkp = &zwork;
  int c;
  short n;
  unsigned short crc;

  if ((c = xhRSin ()) < 0)
    return c;
  zworkp->rtype = c;
  crc = updcrc (c, 0);

  /* read flag and CRC */
  for (n = 4 + 2; --n >= 0; )
    {
      if ((c = xhRSin ()) < 0)
	return c;
      *hdr++ = c;
    }

  crc = mkcrc (hdr - 6, 6, crc);
  if (crc & 0xffff)
    return ERROR;

  /* catch CR and LF */
  c = RSin ();
  if (c == CR || c == (CR | 0x80))
    {
      c = RSin ();
      if (c != LF && c != (LF | 0x80))
	return ERROR;
    }

  return zworkp->rtype;
}

int
zrdata (char *buf, short length)
{
  struct zwork *zworkp = &zwork;
#if !defined (USE_ASM_ZDECODE)
  char *end;
#endif
  int c;
  int terminater;
  unsigned short crc;

  if (zworkp->rcrc32)
    return zrdat32 (buf, length);
  else
    {
#if defined (USE_ASM_ZDECODE)
      c = 0;
      zworkp->rdcount = 0;
      crc = zedmkcrc (&zworkp->rdcount, &c, zworkp->ctlesc, buf, length, 0);
      do
	{
	  switch (c)
	    {
	    case GOTCRCE:
	    case GOTCRCG:
	    case GOTCRCQ:
	    case GOTCRCW:
	      terminater = c;
	      crc = updcrc ((c & 0xff), crc);

	      if ((c = xeRSin ()) & ~0xff)
		continue;
	      crc = updcrc (c, crc);
	      if ((c = xeRSin ()) & ~0xff)
		continue;
	      crc = updcrc (c, crc);
	      if (crc & 0xffff)
		return ERROR;

	      return terminater;
	    case GOTCAN:
	      return ZCAN;
	    case TIMEOUT:
	      return TIMEOUT;
	    default:
	      return c;
	    }
	}
      while (1);
#else
      crc = 0;
      zworkp->rdcount = 0;
      end = buf + length;

      while (buf <= end)
	{
	  if ((c = xeRSin ()) & ~0xff)
	    {
again:
	      switch (c)
		{
		case GOTCRCE:
		case GOTCRCG:
		case GOTCRCQ:
		case GOTCRCW:
		  terminater = c;
		  crc = updcrc ((c & 0xff), crc);

		  if ((c = xeRSin ()) & ~0xff)
		    goto again;
		  crc = updcrc (c, crc);
		  if ((c = xeRSin ()) & ~0xff)
		    goto again;
		  crc = updcrc (c, crc);
		  if (crc & 0xffff)
		    return ERROR;

		  zworkp->rdcount = buf - (end - length);
		  return terminater;
		case GOTCAN:
		  return ZCAN;
		case TIMEOUT:
		  return TIMEOUT;
		default:
		  return c;
		}
	    }
	  *buf++ = c;
	  crc = updcrc (c, crc);
	}
#endif
    }

  return ERROR;
}

int
zrdat32 (char *buf, short length)
{
  struct zwork *zworkp = &zwork;
#if !defined (USE_ASM_ZDECODE)
  char *end;
#endif
  int c;
  int terminater;
  unsigned long crc;

#if defined (USE_ASM_ZDECODE)
  c = 0;
  zworkp->rdcount = 0;
  crc = zedmkcrc32 (&zworkp->rdcount, &c, zworkp->ctlesc, buf, length, 0xffffffffL);

  do
    {
      switch (c)
	{
	case GOTCRCE:
	case GOTCRCG:
	case GOTCRCQ:
	case GOTCRCW:
	  terminater = c;
	  crc = updcrc32 ((c & 0xff), crc);

	  if ((c = xeRSin ()) & ~0xff)
	    continue;
	  crc = updcrc32 (c, crc);
	  if ((c = xeRSin ()) & ~0xff)
	    continue;
	  crc = updcrc32 (c, crc);
	  if ((c = xeRSin ()) & ~0xff)
	    continue;
	  crc = updcrc32 (c, crc);
	  if ((c = xeRSin ()) & ~0xff)
	    continue;
	  crc = updcrc32 (c, crc);
	  if (crc != 0xdebb20e3)
	    return ERROR;

	  return terminater;
	case GOTCAN:
	  return ZCAN;
	case TIMEOUT:
	  return TIMEOUT;
	default:
	  return c;
	}
    }
  while (1);
#else
  crc = 0xffffffffL;
  zworkp->rdcount = 0;
  end = buf + length;

  while (buf <= end)
    {
      if ((c = xeRSin ()) & ~0xff)
	{
again:
	  switch (c)
	    {
	    case GOTCRCE:
	    case GOTCRCG:
	    case GOTCRCQ:
	    case GOTCRCW:
	      terminater = c;
	      crc = updcrc32 ((c & 0xff), crc);

	      if ((c = xeRSin ()) & ~0xff)
		goto again;
	      crc = updcrc32 (c, crc);
	      if ((c = xeRSin ()) & ~0xff)
		goto again;
	      crc = updcrc32 (c, crc);
	      if ((c = xeRSin ()) & ~0xff)
		goto again;
	      crc = updcrc32 (c, crc);
	      if ((c = xeRSin ()) & ~0xff)
		goto again;
	      crc = updcrc32 (c, crc);
	      if (crc != 0xdebb20e3)
		return ERROR;

	      zworkp->rdcount = buf - (end - length);
	      return terminater;
	    case GOTCAN:
	      return ZCAN;
	    case TIMEOUT:
	      return TIMEOUT;
	    default:
	      return c;
	    }
	}
      *buf++ = c;
      crc = updcrc32 (c, crc);
    }
#endif

  return ERROR;
}
