/*
 * Copyright (C) 1995, jack
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

static char rcsid[]= "$Id: zsend.c,v 1.4 1995/11/14 09:14:15 jack Exp $";

#include "config.h"
#include "rsio.h"
#include "updcrc.h"
#include "zsend.h"
#include "zmodem.h"
#include "zwork.h"

#define HEXH(x) hexchar[(((unsigned char) (x)) >> 4) & 0xf]
#define HEXL(x) hexchar[((unsigned char) (x)) & 0xf]

#if defined (USE_ASM_ZENCODE)
extern unsigned short zemkcrc (int *lastsent, int ctlesc, char *buffer,
			       int len, unsigned short crc);
extern unsigned long zemkcrc32 (int *lastsent, int ctrlesc, char *buffer,
				int len, unsigned long crc);
#endif

void
xeRSout (int c)
{
  struct zwork *zworkp = &zwork;

  if (c & 0x60)
    RSout (zworkp->lastsent = c);
  else
    {
      switch (c &= 0xff)
	{
	case ZDLE:
	  RSout (ZDLE);
	  RSout (zworkp->lastsent = (c ^= 0x40));
	  break;
	case CR:
	case CR | 0x80:
	  if (!zworkp->ctlesc && (zworkp->lastsent & 0x7f) != '@')
	    goto sendit;
	case DLE:
	case XON:
	case XOFF:
	case DLE | 0x80:
	case XON | 0x80:
	case XOFF | 0x80:
	  RSout (ZDLE);
	  c ^= 0x40;
sendit:
	  RSout (zworkp->lastsent = c);
	  break;
	default:
	  if (zworkp->ctlesc && ! (c & 0x60))
	    {
	      RSout (ZDLE);
	      c ^= 0x40;
	    }
	  RSout (zworkp->lastsent = c);
	}
    }
}

void
zsbhdr (short len, short type, char *hdr)
{
  struct zwork *zworkp = &zwork;
  short n;
  unsigned short crc;

  if (type == ZDATA)
    for (n = Znulls; --n >= 0; )
      RSout (0);

  RSout (ZPAD);
  RSout (ZDLE);

  if ((zworkp->scrc32 = zworkp->usefcs32))
    zsbh32 (len, type, hdr);
  else
    {
      RSout (ZBIN);
      xeRSout (type);
      crc = updcrc (type, 0);

#if defined (USE_ASM_ZENCODE)
      crc = zemkcrc (&zworkp->lastsent, zworkp->ctlesc, hdr, len, crc);
#else
      crc = mkcrc (hdr, len, crc);
      for (n = len; --n >= 0; )
	xeRSout (*hdr++);
#endif

      crc = updcrc (0, updcrc (0, crc));
      xeRSout (crc >> 8);
      xeRSout (crc);
    }

  if (type != ZDATA)
    RSflush ();
}

void
zsbh32 (short len, short type, char *hdr)
{
  struct zwork *zworkp = &zwork;
#if !defined (USE_ASM_ZENCODE)
  short n;
#endif
  unsigned long crc;

  RSout (ZBIN32);
  xeRSout (type);

  crc = updcrc32 (type, 0xffffffffL);

#if defined (USE_ASM_ZENCODE)
  crc = zemkcrc32 (&zworkp->lastsent, zworkp->ctlesc, hdr, len, crc);
#else
  crc = mkcrc32 (hdr, len, crc);
  for (n = len; --n >= 0; )
    xeRSout (*hdr++);
#endif

  crc = ~crc;
  xeRSout ((int) crc);
  crc >>= 8;
  xeRSout ((int) crc);
  crc >>= 8;
  xeRSout ((int) crc);
  crc >>= 8;
  xeRSout ((int) crc);
}

void
zshhdr (short len, short type, char *hdr)
{
  struct zwork *zworkp = &zwork;
  static char hexchar[16] = "0123456789abcdef";
  short n;
  unsigned short crc;

  RSout (ZPAD);
  RSout (ZPAD);
  RSout (ZDLE);
  RSout (ZHEX);
  zworkp->scrc32 = 0;

  RSout (HEXH (type));
  RSout (HEXL (type));
  crc = updcrc (type, 0);

  crc = mkcrc (hdr, len, crc);
  for (n = len; --n >= 0; ++hdr)
    {
      RSout (HEXH (*hdr));
      RSout (HEXL (*hdr));
    }

  crc = updcrc (0, updcrc (0, crc));
  RSout (HEXH (crc >> 8));
  RSout (HEXL (crc >> 8));
  RSout (HEXH (crc));
  RSout (HEXL (crc));

  RSout (CR);
  RSout (LF);

  if (type != ZFIN && type != ZACK)
    RSout (XON);
  RSflush ();
}

void
zsdata (char *buf, short length, int frameend)
{
  struct zwork *zworkp = &zwork;
  unsigned short crc;

  if (zworkp->scrc32)
    zsdat32 (buf, length, frameend);
  else
    {
#if defined (USE_ASM_ZENCODE)
      crc = zemkcrc (&zworkp->lastsent, zworkp->ctlesc, buf, length, 0);
#else
      crc = mkcrc (buf, length, 0);
      for ( ; --length >= 0; )
	xeRSout (*buf++);
#endif

      RSout (ZDLE);
      RSout (frameend);
      crc = updcrc (frameend, crc);

      crc = updcrc (0, updcrc (0, crc));
      xeRSout (crc >> 8);
      xeRSout (crc);
    }

  if (frameend == ZCRCW)
    RSout (XON);
  if (frameend != ZCRCG)
    RSflush ();
}

void
zsdat32 (char *buf, short length, int frameend)
{
  struct zwork *zworkp = &zwork;
  unsigned long crc;

#if defined (USE_ASM_ZENCODE)
  crc = zemkcrc32 (&zworkp->lastsent, zworkp->ctlesc, buf, length, 0xffffffffL);
#else
  crc = mkcrc32 (buf, length, 0xffffffffL);
  for ( ; --length >= 0; )
    xeRSout (*buf++);
#endif

  RSout (ZDLE);
  RSout (frameend);
  crc = updcrc32 (frameend, crc);

  crc = ~crc;
  xeRSout ((int) crc);
  crc >>= 8;
  xeRSout ((int) crc);
  crc >>= 8;
  xeRSout ((int) crc);
  crc >>= 8;
  xeRSout ((int) crc);
}

void
zcancel (void)
{
  static char cancelstr[] =
  { CAN, CAN, CAN, CAN, CAN, CAN, CAN, CAN, CAN, CAN,
    BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, 0 };

  RSprint (cancelstr);
  RSflush ();
}
